﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Optimization.Operators.LCS {
  [StorableClass]
  [Item("GAssistEnsembleModel", "")]
  public class GAssistEnsembleModel : NamedItem, IGAssistEnsembleModel {

    [Storable]
    private List<IGAssistModel> models;
    public IEnumerable<IGAssistModel> Models {
      get { return new List<IGAssistModel>(models); }
    }

    [StorableConstructor]
    protected GAssistEnsembleModel(bool deserializing) : base(deserializing) { }
    protected GAssistEnsembleModel(GAssistEnsembleModel original, Cloner cloner)
      : base(original, cloner) {
      this.models = original.Models.Select(m => cloner.Clone(m)).ToList();
    }
    public GAssistEnsembleModel() : this(Enumerable.Empty<IGAssistModel>()) { }
    public GAssistEnsembleModel(IEnumerable<IGAssistModel> models)
      : base() {
      this.name = ItemName;
      this.description = ItemDescription;
      this.models = new List<IGAssistModel>(models);
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new GAssistEnsembleModel(this, cloner);
    }

    #region IGAssistEnsembleModel Members
    public void Add(IGAssistModel model) {
      models.Add(model);
    }
    public void Remove(IGAssistModel model) {
      models.Remove(model);
    }

    public IEnumerable<IEnumerable<IGAssistNiche>> GetEstimatedNicheVectors(IEnumerable<IGAssistInput> input) {
      var estimatedValuesEnumerators = (from model in models
                                        select model.Evaluate(input).GetEnumerator())
                                       .ToList();

      while (estimatedValuesEnumerators.All(en => en.MoveNext())) {
        yield return from enumerator in estimatedValuesEnumerators
                     select enumerator.Current;
      }
    }
    #endregion

    #region IGAssistModel Members

    public IEnumerable<IGAssistNiche> Evaluate(IEnumerable<IGAssistInput> input) {
      foreach (var estimatedValuesVector in GetEstimatedNicheVectors(input)) {
        // return the class which is most often occuring
        yield return
          estimatedValuesVector
          .GroupBy(x => x, new GAssistNicheComparer())
          .OrderByDescending(g => g.Count())
          .Select(g => g.Key)
          .First();
      }
    }

    public IGAssistSolution CreateGAssistSolution(IGAssistProblemData problemData) {
      return new GAssistEnsembleSolution(models, new GAssistEnsembleProblemData((IGAssistProblemData)problemData.Clone()));
    }
    #endregion

    public IEnumerable<IGAssistNiche> Evaluate(IEnumerable<IGAssistInput> input, out ItemSet<IItem> aliveRules, out double theoryLength) {
      aliveRules = new ItemSet<IItem>();
      theoryLength = double.NaN;
      return Evaluate(input);
    }
  }
}
